import pya
import cells
from pprint import pprint

from pya import DPoint, DCellInstArray, DTrans, Cell, Layout, DVector, DBox, DPath

import pya
import cells
from pprint import pprint

class KlayoutUtilities:
  _instances = {}

  def __init__(self):
    self.cell_cache = dict()

    self.app: pya.Application = pya.Application.instance()

    self.app.set_config("grid-micron", "0.005")
    #self.app.set_config("grid-micron", "0.001")
    self.app.set_config("grid-style0", "lines")
    

    self.layout_view: pya.LayoutView = self.app.main_window().current_view()
    if self.layout_view is None:
      print("There is any layout open... creating")

      tech="gf180mcu"
      mode=0
      self.app.main_window().create_layout(tech, mode)

      self.layout_view = self.app.main_window().current_view()

    # Cell view can be pointing nothing
    # If that's the case, a TOP cell is created and pointed

    # cell_view is pointing to "TOP"
    self.cell_view: pya.CellView = self.layout_view.active_cellview()

    # layout points to TOP cell layout.
    # We can get cells referenced by TOP
    self.layout: pya.Layout = self.cell_view.layout()

    if self.cell_view.cell is None:
      print("Current cell is not setted (Ctrl-S over top cell)")

      if self.layout.cell("TOP") is None:
        self.layout.create_cell("TOP")
        self.cell_view.set_cell_name("TOP")

    # Direct reference to TOP cell.
    self.viewed_cell: pya.Cell= self.cell_view.cell

  def __call__(cls, *args, **kwargs):
    """Always returns the same KlayoutUtilities instance"""
    if cls not in cls._instances:
      instance = super().__call__(*args, **kwargs)
      cls._instances[cls] = instance

    return cls._instances[cls]

  @staticmethod
  def register(func, *args, **kwargs):
    """Assuming that func is a pure function, the register avoid cell duplication"""
    k = KlayoutUtilities()

    # print(f"{func=}")
    # print(f"{args=}")
    # print(f"{kwargs=}")

    func_key = func.__hash__()
    arg_key = "_".join(str(arg) for arg in args)
    kwarg_key = "_".join(f"{key}_{str(value)}" for key,value in kwargs.items())

    cell_key = f"{func_key}-{arg_key}-{kwarg_key}"

    registered = True
    if not cell_key in k.cell_cache.keys():
      k.cell_cache[cell_key] = func(*args, **kwargs)
      registered = False

    # else:
    #   print("match")

    return k.cell_cache[cell_key], registered


  @staticmethod
  def inject_top_layout(func):
    """Decorator. Injects top cell and the layout"""
    k = KlayoutUtilities()

    def _wrapper(*args, **kwargs):
      cell, _ = k.register(func, *args, layout=k.layout, top=k.viewed_cell, **kwargs)

      return cell
    
    return _wrapper


  @staticmethod
  def clear():
      """Removes all the cells in the hierarchy of the top cell"""
      k = KlayoutUtilities()

      # To clean the top cell, we have to strategies:
      # Delete subcells
      k.layout.prune_subcells(k.viewed_cell.cell_index(), -1)

      # Flatten
      #k.viewed_cell.flatten(-1, prune=True)

      for topcell_idx in k.layout.each_top_cell():
      
          topcell = k.layout.cell(topcell_idx)
          if topcell.name in {"TOP", k.viewed_cell}:
              continue

          k.layout.prune_cell(topcell, -1)

      # This always happens
      k.viewed_cell.clear()
      k.cell_cache.clear()



  @staticmethod
  def set_visual_configuration():
    k = KlayoutUtilities()
    #k.cell_view.set_cell_name("TOP")
    k.layout_view.add_missing_layers()
    k.layout_view.zoom_fit()


fet_3p3_l = float(0.28)
fet_3p3_w = float(0.22)
fet_w_con = float(0.36)
fet_w_con_bulk = 0.42
fet_5_6_w = 0.3

nfet_05v0_l = 0.6
nfet_06v0_l = 0.7

pfet_05v0_l = 0.5
pfet_06v0_l = 0.55

nfet_nat_l = 1.8
nfet_nat_w = 0.8
fet_grw = 0.38
fet_ld = 0.52

ldfet_l_min = 0.6
ldfet_l_max = 20
ldfet_w_min = 4
ldfet_w_max = 50


def clean_pmos_parameters(pmos_params) -> dict[str, int | float | str]:
    "Internal funtion to avoiding errors in pmos generations"

    # Making a copy
    pmos_params = dict(pmos_params)

    #pmos_params["area"] = pmos_params["w_gate"] * pmos_params["l_gate"]
    #pmos_params["perim"] = 2 * (pmos_params["w_gate"] + pmos_params["l_gate"])
    
    # w,l must be larger or equal than min. values.
    if pmos_params["volt"] == "3.3V":
        if (pmos_params["l_gate"]) < fet_3p3_l:
            pmos_params["l_gate"] = fet_3p3_l
        if (pmos_params["w_gate"]) < fet_3p3_w:
            pmos_params["w_gate"] = fet_3p3_w

    elif pmos_params["volt"] == "5V":
        if (pmos_params["l_gate"]) < pfet_05v0_l:
            pmos_params["l_gate"] = pfet_05v0_l
        if (pmos_params["w_gate"]) < fet_5_6_w:
            pmos_params["w_gate"] = fet_5_6_w

    elif pmos_params["volt"] == "6V":
        if (pmos_params["l_gate"]) < pfet_06v0_l:
            pmos_params["l_gate"] = pfet_06v0_l
        if (pmos_params["w_gate"]) < fet_5_6_w:
            pmos_params["w_gate"] = fet_5_6_w

    if (pmos_params["inter_sd_l"]) < fet_ld:
        pmos_params["inter_sd_l"] = fet_ld

    if (pmos_params["grw"]) < fet_grw:
        pmos_params["grw"] = fet_grw

    if pmos_params["con_bet_fin"] == 1:
        if pmos_params["bulk"] == "Bulk Tie":
            if (pmos_params["w_gate"]) < fet_w_con_bulk:
                pmos_params["w_gate"] = fet_w_con_bulk
        else:
            if (pmos_params["w_gate"]) < fet_w_con:
                pmos_params["w_gate"] = fet_w_con

    return pmos_params


def default_pmos_parameters():
    return dict({
        "layout": None,                 # layout
        "deepnwell": 0,                 # Deep NWELL
        "pcmpgr": 0,                    # Deep NWELL Guard Ring
        "volt": "3.3V",                 # Operating Voltage (3.3V 5V 6V)
        "bulk": "None",                 # Bulk Type (None, Bulk Tie, Guard Ring)
        "w_gate": fet_3p3_w,            # Width
        "l_gate": fet_3p3_l,            # Length
        "inter_sd_l": fet_ld,           # Diffusion Length (ld)
        "nf": 1,                        # Number of Fingers
        "grw": fet_grw,                 # Guard Ring Width
        "gate_con_pos": "bottom",  # Gate Contact Position (top bottom alternating)
        "con_bet_fin": 1,               # Contact Between Fingers
        "sd_con_col": 1,                # Diffusion Contacts Columns
        "interdig": 0,                  # Interdigitation
        "patt": "",                     # Pattern in case of Interdigitation
        "patt_lbl": 0,                  # Interdigitation pattern label
        "lbl": 0,                       # Labels
        "sd_lbl": list(),               # Pattern of Source/Drain Labels
        "g_lbl": list(),                # Patterns of Gate Labels
        "sub_lbl": "",                  # Substrate Label
    })


def pmos_dimentions(l, w, nf) -> dict:
    pmos_params = default_pmos_parameters()

    # Calculate x length

    pmos_dim = dict()

    # Calculate x length

    pmos_dim["total_channel_length"] = l * nf
    pmos_dim["total_channel_width"] = w / nf

    pmos_dim["diffusion_length"] = pmos_params["inter_sd_l"]
    pmos_dim["total_diffusion_length"] = (nf-1)*pmos_dim["diffusion_length"]

    pmos_dim["drain_source_length_offset"] = 0.13
    
    pmos_dim["drain_length"] = 0.65 #0.46 + 0.19
    pmos_dim["source_length"] = 0.65 #0.46 + 0.19
    pmos_dim["bulk_length"] = 0.73 # 0.21 + 0.52

    pmos_dim["x_offset"] = -0.46
    pmos_dim["x_length"] = (
        pmos_dim["total_channel_length"] +
        pmos_dim["total_diffusion_length"] + 
        pmos_dim["drain_length"] + 
        pmos_dim["source_length"] + 
        pmos_dim["bulk_length"]
    )
    
    pmos_dim["y_offset"] = 0
    pmos_dim["y_length"] = (
        pmos_dim["total_channel_width"]
    )

    pmos_dim["first_poly_contact_center_x"] = 0.19 + l/2
    pmos_dim["first_poly_contact_center_y"] = pmos_dim["total_channel_width"] + 0.49

    return pmos_dim


# PMOS design
#############

@KlayoutUtilities.inject_top_layout
def generate_pmos(position, l, w, nf=1, top: Cell = None, layout: Layout = None):
    pmos_params: dict = default_pmos_parameters()
    pmos_params["layout"] = layout

    pmos_params["bulk"] = "None"
    pmos_params["w_gate"] = w / nf
    pmos_params["l_gate"] = l
    pmos_params["nf"] = nf

    pmos_params["gate_con_pos"] = "top"
    pmos_params["con_bet_fin"] = 1
    pmos_params["sd_con_col"] = 1

    pmos_params = clean_pmos_parameters(pmos_params)

    pmos: Cell = cells.draw_fet.draw_pfet(
        **pmos_params 
    )
    #pmos.flatten(1)

    pmos_inst = DCellInstArray(
        pmos,
        DTrans(position), # 0.32300
        DVector(0, 0),
        DVector(0, 0),
        1,
        1,
    )

    top.insert(pmos_inst)

    return pmos_inst


KlayoutUtilities.clear()

l=0.7
w=5
nf=1

generate_pmos(DPoint(0, 0.000), l, w, nf)

KlayoutUtilities.set_visual_configuration()